import datetime
import logging
from typing import List, Dict

from influxdb_client import InfluxDBClient, Point, WriteOptions, WriteApi, QueryApi, WritePrecision, BucketsApi
from influxdb_client.client.delete_api import DeleteApi
from influxdb_client.client.flux_table import FluxRecord
from influxdb_client.client.write_api import WriteType

from galileodb import ExperimentDatabase, Experiment, NodeInfo, Telemetry
from galileodb.model import ExperimentEvent, RequestTrace

logger = logging.getLogger()


class InfluxExperimentDatabase(ExperimentDatabase):
    client: InfluxDBClient
    writer: WriteApi
    query: QueryApi
    delete: DeleteApi
    bucket: BucketsApi

    def __init__(self, client: InfluxDBClient, org_name: str = 'galileo', org_id='org-id') -> None:
        super().__init__()
        self.client = client
        self.org_name = org_name
        self.org_id = org_id
        self.write_options = WriteOptions(write_type=WriteType.synchronous)

    def open(self):
        self.writer = self.client.write_api(self.write_options)
        self.query = self.client.query_api()
        self.bucket = self.client.buckets_api()
        self.delete = self.client.delete_api()

    def close(self):
        self.client.close()

    def save_experiment(self, experiment: Experiment):
        raise NotImplementedError()

    def update_experiment(self, experiment: Experiment):
        raise NotImplementedError()

    def delete_experiment(self, exp_id: str):
        exp_bucket = self.bucket.find_bucket_by_name(exp_id)
        if exp_bucket is None:
            logger.info(f'Did not find bucket for experiment with ID: {exp_id}')
        else:
            self.bucket.delete_bucket(exp_bucket)

    def get_experiment(self, exp_id: str) -> Experiment:
        raise NotImplementedError()

    def touch_traces(self, experiment: Experiment):
        raise NotImplementedError()

    def save_metadata(self, exp_id: str, data: Dict):
        raise NotImplementedError()

    def get_metadata(self, exp_id: str) -> Dict:
        raise NotImplementedError()

    @staticmethod
    def _map_flux_record_to_request_trace(record: FluxRecord) -> RequestTrace:
        return RequestTrace(
            request_id=record.get_value(),
            client=record.values['client'],
            service=record.values['service'],
            done=float(record.values['done']),
            created=float(record.values['created']),
            sent=float(record.values['sent']),
            status=int(record.values['status']),
            server=record.values['server'],
            exp_id=record.values['exp_id'],
            headers=record.values.get('headers', None),
            response=record.values['response']
        )

    def get_traces(self, exp_id: str) -> List[RequestTrace]:
        records = self._query_for_measurement("traces", exp_id)
        events = list()

        for record in records:
            events.append(self._map_flux_record_to_request_trace(record))

        return events

    # https://github.com/influxdata/influxdb-client-python/blob/eadbf6ac014582127e2df54698682e2924973e19/examples/nanosecond_precision.py#L37

    def get_telemetry(self, exp_id=None) -> List[Telemetry]:
        records = self._query_for_measurement("telemetry", exp_id)
        events = list()

        for record in records:
            events.append(self._map_flux_record_to_telemetry(record))

        return events

    @staticmethod
    def _map_flux_record_to_telemetry(record: FluxRecord):
        return Telemetry(
            timestamp=float(record.values['ts']),
            metric=record.values['metric'],
            node=record.values['node'],
            value=record.get_value(),
            exp_id=record.values['exp_id'],
            subsystem=record.values.get('subsystem')
        )

    def save_telemetry(self, telemetry: List[Telemetry]):
        if len(telemetry) == 0:
            return

        points: List[Point] = list()
        for data in telemetry:
            strftime = datetime.datetime.utcfromtimestamp(float(data.timestamp))
            p = Point("telemetry") \
                .time(strftime, WritePrecision.MS) \
                .tag('ts', float(data.timestamp)) \
                .field('value', float(data.value)) \
                .tag('exp_id', data.exp_id) \
                .tag('node', data.node) \
                .tag('metric', data.metric) \
                .tag('subsystem', data.subsystem)
            points.append(p)

        self.writer.write(bucket=telemetry[0].exp_id, org=self.org_name, record=points)

    def save_event(self, event: ExperimentEvent):
        return self.save_events([event])

    def save_traces(self, traces: List[RequestTrace]):
        if len(traces) == 0:
            return

        points: List[Point] = list()
        for trace in traces:
            strftime = datetime.datetime.utcfromtimestamp(float(trace.sent))
            p = Point("traces") \
                .time(strftime, WritePrecision.MS) \
                .field("request_id", trace.request_id) \
                .tag("client", trace.client) \
                .tag("service", trace.service) \
                .tag("created", float(trace.created)) \
                .tag("sent", float(trace.sent)) \
                .tag("done", float(trace.done)) \
                .tag("status", int(trace.status)) \
                .tag("server", trace.server) \
                .tag("exp_id", trace.exp_id) \
                .tag("response", trace.response) \
                .tag("headers", trace.headers)
            points.append(p)

        self.writer.write(bucket=traces[0].exp_id, org=self.org_name, record=points)

    def save_events(self, events: List[ExperimentEvent]):
        if len(events) == 0:
            return

        points: List[Point] = list()

        for event in events:
            strftime = datetime.datetime.utcfromtimestamp(float(event.timestamp))
            p = Point("events") \
                .time(strftime, WritePrecision.MS) \
                .tag("ts", float(event.timestamp)) \
                .tag("name", event.name) \
                .field("value", event.value) \
                .tag("exp_id", event.exp_id)
            points.append(p)

        self.writer.write(bucket=events[0].exp_id, org=self.org_name, record=points)

    @staticmethod
    def _map_flux_record_to_exp_event(record: FluxRecord) -> ExperimentEvent:
        return ExperimentEvent(
            exp_id=record.values['exp_id'],
            timestamp=float(record.values['ts']),
            name=record.values['name'],
            value=record.get_value()
        )

    def _query_for_measurement(self, measurement: str, exp_id: str):
        stop = datetime.datetime.utcnow() + datetime.timedelta(days=1)
        stop = stop.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        records = self.query.query_stream(
            f'''  
               from(bucket: "{exp_id}")
                 |> range(start: 1970-01-01, stop: {stop})
                 |> filter(fn: (r) => r["_measurement"] == "{measurement}")    
            '''
        )
        return records

    def get_events(self, exp_id) -> List[ExperimentEvent]:
        records = self._query_for_measurement("events", exp_id)
        events = []
        for record in records:
            events.append(InfluxExperimentDatabase._map_flux_record_to_exp_event(record))

        return events

    def save_nodeinfos(self, infos: List[NodeInfo]):
        raise NotImplementedError()

    def find_all(self) -> List[Experiment]:
        raise NotImplementedError()

    def get_running_experiment(self) -> Experiment:
        raise NotImplementedError()



import os
import tempfile
import unittest

from galileodb.sql.driver.sqlite import SqliteAdapter
from tests.sql.adapter import AbstractTestSqlDatabase


class TestSqliteDatabase(AbstractTestSqlDatabase, unittest.TestCase):
    db_file = None

    def setUp(self) -> None:
        self.db_file = tempfile.mktemp('.sqlite', 'galileo_test_')
        super().setUp()

    def _create_sql_adapter(self):
        return SqliteAdapter(self.db_file)

    def tearDown(self) -> None:
        super().tearDown()
        os.remove(self.db_file)


if __name__ == '__main__':
    unittest.main()



from typing import Iterable

from galileodb.model import RequestTrace


class RedisTraceReporter:
    channel = 'galileo/results/traces'
    line_format = '%s,%s,%s,%.7f,%.7f,%.7f,%d,%s,%s,%s,%s'

    def __init__(self, rds) -> None:
        super().__init__()
        self.rds = rds

    def report_multiple(self, traces: Iterable[RequestTrace]):
        rds = self.rds.pipeline()
        key = self.channel
        fmt = self.line_format

        for t in traces:
            # FIXME: this is turning into a bad line-based protocol ...
            response = t.response
            if response:
                response = response.replace('\n', '\\n')
            headers = t.headers
            if headers is None:
                headers = "None"
            else:
                headers = headers.replace(",", "|")
            value = fmt % (
                t.request_id,
                t.client,
                t.service,
                t.created,
                t.sent,
                t.done,
                t.status,
                t.server,
                t.exp_id,
                headers,
                response
            )
            rds.publish(key, value)

        rds.execute()



import uuid
from datetime import datetime
from typing import NamedTuple, List, Dict


def generate_experiment_id():
    prefix = datetime.strftime(datetime.now(), '%Y%m%d%H%M')
    suffix = str(uuid.uuid4())[:4]
    return prefix + '-' + suffix


class Experiment:
    id: str
    name: str
    creator: str
    start: float
    end: float
    created: float
    status: str

    def __init__(self, id: str = None, name: str = None, creator: str = None, start: float = None, end: float = None,
                 created: float = None, status: str = None) -> None:
        super().__init__()
        self.id = id
        self.name = name
        self.creator = creator
        self.start = start
        self.end = end
        self.created = created
        self.status = status

    def __repr__(self):
        return self.__str__()

    def __str__(self) -> str:
        return 'Experiment%s' % self.__dict__


class Telemetry(NamedTuple):
    timestamp: float
    metric: str
    node: str
    value: float
    exp_id: str
    subsystem: str = None


class NodeInfo(NamedTuple):
    node: str
    data: Dict[str, str]
    exp_id: str


class Event(NamedTuple):
    timestamp: float
    name: str
    value: str = None


class ExperimentEvent(NamedTuple):
    exp_id: str
    timestamp: float
    name: str
    value: str = None


class WorkloadConfiguration(NamedTuple):
    service: str
    ticks: List[int]
    clients_per_host: int
    arrival_pattern: str
    client: str = None
    client_parameters: dict = dict()


class ExperimentConfiguration(NamedTuple):
    duration: int
    interval: int
    workloads: List[WorkloadConfiguration]


class QueuedExperiment(NamedTuple):
    experiment: Experiment
    configuration: ExperimentConfiguration


class RequestTrace(NamedTuple):
    request_id: str
    client: str
    service: str
    created: float
    sent: float
    done: float
    status: int = -1
    server: str = None
    exp_id: str = None
    headers: str = None
    response: str = None  # should be the last field for easier line-based transmission



import argparse
from datetime import datetime
from galileodb.factory import create_experiment_database_from_env


def delete_exp(args):
    print(f'deleting {args.exp_id}...')
    exp_db = create_experiment_database_from_env()
    exp_db.open()

    try:
        exp_db.delete_experiment(args.exp_id)
        print(f'deleted {args.exp_id}')
    except ValueError:
        print(f'could not find experiment {args.exp_id}')


def show_exp(args):
    print(f'showing {args.exp_id}...')
    exp_db = create_experiment_database_from_env()
    exp_db.open()

    exp = exp_db.get_experiment(args.exp_id)

    if exp:
        print(f'experiment: {exp}')
    else:
        print(f'could not find experiment {args.exp_id}')


def list_exp(args):
    exp_db = create_experiment_database_from_env()
    exp_db.open()

    exps = exp_db.find_all()

    if not exps:
        print(f'no experiments found')

    exps = sorted(exps, key=lambda e: e.created)

    for exp in exps:
        if len(exp.name) > 30:
            name = str(exp.name)[:26] + ' ...'
        else:
            name = exp.name

        created = datetime.fromtimestamp(exp.created)

        line = f'| {exp.id:20s} | {name:31s} | {exp.creator:23s} | {created:%d, %b %Y %H:%M} |'
        print(line)

    print('+' + ('-' * (len(line) - 2)) + '+')


def get_running_experiment_id(args):
    exp_db = create_experiment_database_from_env()
    exp_db.open()

    exp = exp_db.get_running_experiment()

    if exp:
        print(f'{exp.id}')
    else:
        print()


def main():
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers()
    sp_delete = sp.add_parser('delete', help='delete the experiment')
    sp_show = sp.add_parser('show', help='show the experiment')
    sp_list = sp.add_parser('list', help='list all experiments')
    parser.add_argument('--exp_id', help='the id of the experiment', required=False)

    sp_running_exp = sp.add_parser('get_running_exp_id', help='get the id of the currently running experiment')

    sp_delete.set_defaults(func=delete_exp)
    sp_show.set_defaults(func=show_exp)
    sp_list.set_defaults(func=list_exp)
    sp_running_exp.set_defaults(func=get_running_experiment_id)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



import argparse
import logging
import os
import signal
import time

import redis

from galileodb.factory import create_experiment_database_from_env
from galileodb.model import Experiment, generate_experiment_id
from galileodb.recorder import Recorder

logger = logging.getLogger(__name__)


def create_experiment(args):
    experiment_id = generate_experiment_id()
    if args.name:
        name = args.name
    else:
        name = experiment_id

    if args.creator:
        creator = args.creator
    else:
        creator = 'galileodb-recorder-' + str(os.getpid())

    now = time.time()

    return Experiment(experiment_id, name=name, creator=creator, start=now, created=now, status='RUNNING')


def create_redis():
    host = os.getenv('galileo_redis_host', 'localhost')
    password = os.getenv('galileo_redis_password', None)
    port = int(os.getenv('galileo_redis_port', 6379))

    logger.info('connecting to redis event bus on %s:%d', host, port)

    return redis.Redis(host=host, port=port, password=password, decode_responses=True)


def save_metadata(exp_id: str, args, exp_db):
    if args.metadata:
        metadata = args.metadata
    else:
        metadata = {}

    exp_db.save_metadata(exp_id, metadata)


def run(args):
    signal.signal(signal.SIGTERM, handle_sigterm)

    # connect to redis eventbus
    rds = create_redis()

    # connect to experiment database
    exp_db = create_experiment_database_from_env()
    exp_db.open()

    # create and save the experiment
    exp = create_experiment(args)
    exp_db.save_experiment(exp)

    # save metadata
    exp_id = exp.id
    save_metadata(exp_id, args, exp_db)

    # main control loop
    recorder = Recorder(rds, exp_db, exp.id)
    try:
        logger.info('starting experiment recorder for exp %s', exp.id)
        recorder.start()
        logger.debug('storing node info keys')
        recorder.telemetry_recorder.save_nodeinfos()
        recorder.join()
    except KeyboardInterrupt:
        logger.debug('interrupt received')
        pass
    finally:
        exp_db.finalize_experiment(exp, 'FINISHED')
        logger.info('shutting down experiment recorder')
        if recorder:
            try:
                recorder.stop(5)
            except:
                pass

        exp_db.close()

    logger.info('experiment %s exiting', exp.id)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--name', required=False, help='set name of experiment', default='')
    parser.add_argument('--creator', required=False, help='set name of creator', default='')
    args = parser.parse_args()

    logging.basicConfig(level=logging._nameToLevel[os.getenv('galileo_log_level', 'INFO')])

    run(args)


def handle_sigterm(signal_number, _stack_frame):
    logger.debug('received signal %s', signal_number)
    raise KeyboardInterrupt


if __name__ == '__main__':
    main()



import logging
import threading
from abc import ABC
from typing import Iterator

from galileodb.model import RequestTrace
from galileodb.reporter.traces import RedisTraceReporter
from galileodb.trace import TraceWriter

logger = logging.getLogger(__name__)


class TraceRecorder(threading.Thread, ABC):

    def __init__(self, rds, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.rds = rds
        self._sub = None

    def stop(self, timeout=None):
        """
        Equivalent to recorder.close() and recorder.join(timeout).

        :param timeout: the join timeout
        """
        self.close()
        self.join(timeout=timeout)

    def close(self):
        if self._sub:
            self._sub.close()

    def run(self):
        self._sub = TracesSubscriber(self.rds)
        sub = self._sub.run()

        try:
            for trace in sub:
                self._record(trace)
        finally:
            self._sub.close()

    def _record(self, service_request: RequestTrace):
        raise NotImplementedError


class RedisTraceRecorder(TraceRecorder):

    def __init__(self, rds, exp_id: str, writer: TraceWriter, flush_every=36) -> None:
        super().__init__(rds)
        self.exp_id = exp_id
        self.writer = writer
        self.buffer = list()

        self.flush_every = flush_every
        self.i = 0

    def run(self):
        try:
            logger.debug('starting RedisTraceRecorder for experiment %s', self.exp_id)
            super().run()
        finally:
            logger.debug('closing RedisTraceRecorder for experiment %s', self.exp_id)
            self._flush()

    def _record(self, t: RequestTrace):
        t = t._replace(exp_id=self.exp_id)
        self.buffer.append(t)

        self.i = (self.i + 1) % self.flush_every
        if self.i == 0:
            self._flush()

    def _flush(self):
        self.writer.write(self.buffer)
        self.buffer.clear()


class TracesSubscriber:
    def __init__(self, rds, channel=None) -> None:
        super().__init__()
        self.rds = rds
        self.channel = channel or RedisTraceReporter.channel
        self.line_format = RedisTraceReporter.line_format
        self.pubsub = None

    def run(self) -> Iterator[RequestTrace]:
        self.pubsub = self.rds.pubsub()

        try:
            self.pubsub.subscribe(self.channel)

            for item in self.pubsub.listen():
                data = item['data']
                if type(data) == int:
                    continue
                try:
                    yield self.parse(data)
                except Exception as e:
                    logger.error('error parsing data string `%s`: %s', data, e)
        finally:
            self.pubsub.close()

    def close(self):
        if self.pubsub:
            self.pubsub.unsubscribe()

    @staticmethod
    def parse(line: str) -> RequestTrace:
        # FIXME: this is turning into a bad line-based protocol ...
        data = line.split(',', maxsplit=10)

        for i in range(len(data)):
            if data[i] == 'None':
                data[i] = None

        response = data[10]

        if response:
            response = response.replace('\\n', '\n')
        return RequestTrace(
            request_id=data[0],
            client=data[1],
            service=data[2],
            created=float(data[3]),
            sent=float(data[4]),
            done=float(data[5]),
            status=int(data[6]),
            server=data[7],
            exp_id=data[8],
            headers=data[9].replace("|", ",") if data[9] is not None else None,
            response=response
        )



from telemc import Telemetry


class RedisTelemetryReporter:

    # TODO belongs in telemc

    def __init__(self, rds) -> None:
        self.rds = rds

    def report(self, telemetry: Telemetry):
        channel = f'telem/{telemetry.node}/{telemetry.metric}'

        if telemetry.subsystem:
            channel += '/' + telemetry.subsystem

        msg = f'{telemetry.timestamp} {telemetry.value}'

        self.rds.publish(channel, msg)



from galileodb.recorder.events import ExperimentEventRecorder, ExperimentEventRecorderThread
from galileodb.recorder.recorder import Recorder
from galileodb.recorder.telemetry import ExperimentTelemetryRecorder

name = 'recorder'

__all__ = [
    'Recorder',
    'ExperimentEventRecorder',
    'ExperimentEventRecorderThread',
    'ExperimentTelemetryRecorder'
]



import sqlite3

from galileodb.sql.adapter import SqlAdapter


class SqliteAdapter(SqlAdapter):
    placeholder = '?'

    def _connect(self, *args, **kwargs):
        return sqlite3.connect(*args, **kwargs)



import abc
import json
import logging
import os
import threading
from typing import Tuple, List, Dict, Optional

from galileodb.db import ExperimentDatabase
from galileodb.model import Experiment, Telemetry, RequestTrace, NodeInfo, ExperimentEvent

logger = logging.getLogger(__name__)


class SqlAdapter(abc.ABC):
    placeholder = '?'

    _thread_local = threading.local()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._thread_local.connection = None
        self.connect_args = args
        self.connect_kwargs = kwargs
        self._lock = threading.RLock()

    @property
    def connection(self):
        with self._lock:
            if 'connection' not in self._thread_local.__dict__ or self._thread_local.connection is None:
                logger.info('%s connecting to database', threading.current_thread().name)
                self._thread_local.connection = self._connect(*self.connect_args, **self.connect_kwargs)
            return self._thread_local.connection

    def reconnect(self):
        with self._lock:
            self.close()
            self.open()

    @property
    def db(self):
        return self.connection

    def cursor(self):
        return self.db.cursor()

    def execute(self, *args, **kwargs):
        cur = self.cursor()
        try:
            logger.debug('executing SQL %s %s', args, kwargs)
            cur.execute(*args, **kwargs)
            self.db.commit()
        finally:
            cur.close()

    def executemany(self, *args, **kwargs):
        cur = self.cursor()
        try:
            cur.executemany(*args, **kwargs)
            self.db.commit()
        finally:
            cur.close()

    def executescript(self, *args, **kwargs):
        cur = self.cursor()
        try:
            cur.executescript(*args, **kwargs)
            self.db.commit()
        finally:
            cur.close()

    def fetchone(self, *args, **kwargs):
        cur = self.cursor()
        try:
            cur.execute(*args, **kwargs)
            return cur.fetchone()
        finally:
            cur.close()

    def fetchmany(self, *args, **kwargs):
        cur = self.cursor()
        try:
            cur.execute(*args, **kwargs)
            return cur.fetchmany()
        finally:
            cur.close()

    def fetchall(self, *args, **kwargs):
        cur = self.cursor()
        try:
            cur.execute(*args, **kwargs)
            return cur.fetchall()
        finally:
            cur.close()

    def open(self):
        with self._lock:
            assert self.connection is not None

    def close(self):
        with self._lock:
            if 'connection' in self._thread_local.__dict__ and self._thread_local.connection is not None:
                try:
                    self._thread_local.connection.close()
                finally:
                    self._thread_local.connection = None

    def insert_one(self, table: str, data: Dict[str, object]):
        columns = self.sql_field_list(data.keys())
        placeholders = ','.join([self.placeholder] * len(data))
        values = list(data.values())

        # TODO: sanitize table and column inputs
        sql = f'INSERT INTO `{table}` ({columns}) VALUES ({placeholders})'

        logger.debug('running insert sql: %s' % sql)

        self.execute(sql, values)

    def insert_many(self, table: str, keys, data: List):
        columns = self.sql_field_list(keys)
        placeholders = ','.join([self.placeholder] * len(keys))

        # TODO: sanitize table and column inputs
        sql = f'INSERT INTO `{table}` ({columns}) VALUES ({placeholders})'

        logger.debug('running insert many sql on %d items: %s', len(data), sql)

        self.executemany(sql, data)

    def update_by_id(self, table: str, identity: Tuple[str, object], data: Dict[str, object]):
        set_statements, values = list(), list()
        id_col, id_val = identity

        for key, value in data.items():
            set_statements.append('`%s` = %s' % (key.upper(), self.placeholder))
            values.append(value)

        values.append(id_val)

        # TODO: sanitize table and column inputs
        sql = 'UPDATE `{table}` SET {set_statements} WHERE `{id_col}` = ' + self.placeholder
        sql = sql.format(table=table, set_statements=','.join(set_statements), id_col=id_col)

        logger.debug('running update sql: %s', sql)
        self.execute(sql, values)

    def sql_field_list(self, fields, table_prefix: str = None, uppercase=True) -> str:
        return ', '.join([self.sql_field_name(field, table_prefix, uppercase) for field in fields])

    def sql_field_name(self, field, table_prefix: str = None, uppercase=True):
        f = field.upper() if uppercase else field
        if table_prefix:
            return f'`{table_prefix}`.`{f}`'
        else:
            return f'`{f}`'

    def _connect(self, *args, **kwargs):
        raise NotImplementedError


class ExperimentSQLDatabase(ExperimentDatabase):

    SCHEMA_FILE = os.path.join(os.path.dirname(__file__), 'schema.sql')

    def __init__(self, db: SqlAdapter) -> None:
        super().__init__()
        self.db = db

    def read_schema_file(self):
        with open(self.SCHEMA_FILE, 'r') as fd:
            return fd.read()

    def open(self):
        self.db.open()
        self.db.executescript(self.read_schema_file())

    def close(self):
        self.db.close()

    def save_experiment(self, experiment: Experiment):
        data = dict(experiment.__dict__)
        data['exp_id'] = data['id']
        del data['id']
        logger.debug('saving experiment with data %s', data)
        self.db.insert_one('experiments', data)

    def save_metadata(self, exp_id: str, data: Dict):
        logger.debug(f"saving metadata for exp: {exp_id}")
        data['exp_id'] = exp_id
        to_save = {
            'exp_id' : exp_id,
            'data': json.dumps(data)
        }
        self.db.insert_one('metadata', to_save)

    def get_metadata(self, exp_id: str) -> Dict:
        logger.debug(f"get metadata for exp_id: {exp_id}")
        sql = f'SELECT * FROM `metadata` WHERE EXP_ID={self.db.placeholder}'
        entry = self.db.fetchone(sql, (exp_id,))

        if entry:
            row = tuple(entry)
            data = row[1]
            return json.loads(data)
        else:
            return None


    def update_experiment(self, experiment: Experiment):
        data = dict(experiment.__dict__)
        del data['id']

        self.db.update_by_id('experiments', ('exp_id', experiment.id), data)

    def delete_experiment(self, exp_id: str):
        experiment = self.get_experiment(exp_id)
        if experiment is None:
            raise ValueError('No such experiment %s' % exp_id)

        stmts = [
            # "DELETE FROM `telemetry` WHERE EXP_ID = " + self.db.placeholder,
            # "DELETE FROM `traces` WHERE EXP_ID = " + self.db.placeholder,
            # "DELETE FROM `events` WHERE EXP_ID = " + self.db.placeholder,
            "DELETE FROM `experiments` WHERE EXP_ID = " + self.db.placeholder,
            "DELETE FROM `metadata` WHERE EXP_ID = " + self.db.placeholder,
        ]

        for sql in stmts:
            try:
                self.db.execute(sql, (exp_id,))
            except Exception as e:
                logger.exception('Exception while executing %s', sql, e)

    def get_experiment(self, exp_id: str) -> Experiment:
        sql = f'SELECT * FROM `experiments` WHERE EXP_ID = {self.db.placeholder}'

        entry = self.db.fetchone(sql, (exp_id,))

        if entry:
            row = tuple(entry)
            return Experiment(*row)
        else:
            return None

    def save_traces(self, traces: List[RequestTrace]):
        self.db.insert_many('traces', RequestTrace._fields, traces)

    def touch_traces(self, experiment: Experiment):
        sql = 'UPDATE `traces` SET `EXP_ID` = ? WHERE CREATED >= ? AND CREATED <= ?'
        sql = sql.replace('?', self.db.placeholder)
        self.db.execute(sql, (experiment.id, experiment.start, experiment.end))

    def get_traces(self, exp_id=None) -> List[RequestTrace]:
        fields = self.db.sql_field_list(RequestTrace._fields)

        if exp_id is None:
            sql = f'SELECT {fields} from `traces`'
            entries = self.db.fetchall(sql)
        else:
            sql = f'SELECT {fields} from `traces` WHERE EXP_ID = {self.db.placeholder}'
            entries = self.db.fetchall(sql, (exp_id,))

        return list(map(lambda x: RequestTrace(*(tuple(x))), entries))

    def save_telemetry(self, telemetry: List[Telemetry]):
        self.db.insert_many('telemetry', Telemetry._fields, telemetry)

    def get_telemetry(self, exp_id=None) -> List[Telemetry]:
        fields = self.db.sql_field_list(Telemetry._fields)

        if exp_id is None:
            sql = f'SELECT {fields} FROM `telemetry`'
            entries = self.db.fetchall(sql)
        else:
            sql = f'SELECT {fields} FROM `telemetry` WHERE EXP_ID = {self.db.placeholder}'
            entries = self.db.fetchall(sql, (exp_id,))

        return list(map(lambda x: Telemetry(*(tuple(x))), entries))

    def save_event(self, event: ExperimentEvent):
        self.db.insert_one('events', event._asdict())

    def save_events(self, events: List[ExperimentEvent]):
        self.db.insert_many('events', ExperimentEvent._fields, events)

    def get_events(self, exp_id=None) -> List[ExperimentEvent]:
        fields = self.db.sql_field_list(ExperimentEvent._fields)

        if exp_id is None:
            sql = f'SELECT {fields} FROM `events`'
            entries = self.db.fetchall(sql)
        else:
            sql = f'SELECT {fields} FROM `events` WHERE EXP_ID = {self.db.placeholder}'
            entries = self.db.fetchall(sql, (exp_id,))

        return list(map(lambda x: ExperimentEvent(*(tuple(x))), entries))

    def save_nodeinfos(self, infos: List[NodeInfo]):
        keys = ('exp_id', 'node', 'info_key', 'info_value')

        data = list()
        for info in infos:
            for k, v in info.data.items():
                data.append((info.exp_id, info.node, k, v))

        self.db.insert_many('nodeinfo', keys, data)

    def find_all(self) -> List[Experiment]:
        fields = ['exp_id', 'name', 'creator', 'start', 'end', 'created', 'status']
        fields = self.db.sql_field_list(fields)

        sql = f'SELECT {fields} FROM `experiments`'

        entries = self.db.fetchall(sql)

        return list(map(lambda x: Experiment(*(tuple(x))), entries))

    def get_running_experiment(self) -> Optional[Experiment]:
        fields = ['exp_id', 'name', 'creator', 'start', 'end', 'created', 'status']
        fields = self.db.sql_field_list(fields)
        sql = f"SELECT {fields} FROM `experiments` WHERE `STATUS` = 'RUNNING'"

        entry = self.db.fetchone(sql)

        if entry:
            row = tuple(entry)
            return Experiment(*row)
        else:
            return None



import time
from abc import ABC
from typing import List, Dict

from galileodb.model import Experiment, Telemetry, NodeInfo, ExperimentEvent, RequestTrace


class ExperimentDatabase(ABC):

    def open(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def save_experiment(self, experiment: Experiment):
        raise NotImplementedError

    def update_experiment(self, experiment: Experiment):
        raise NotImplementedError

    def delete_experiment(self, exp_id: str):
        raise NotImplementedError

    def get_experiment(self, exp_id: str) -> Experiment:
        raise NotImplementedError

    def save_metadata(self, exp_id: str, data: Dict):
        raise NotImplementedError

    def get_metadata(self, exp_id: str) -> Dict:
        raise NotImplementedError

    def save_traces(self, traces: List[RequestTrace]):
        """
        Saves multiple RequestTrace tuples.
        :param traces: a list of RequestTrace tuples
        :return:
        """
        raise NotImplementedError

    def touch_traces(self, experiment: Experiment):
        raise NotImplementedError

    def get_traces(self, exp_id: str) -> List[RequestTrace]:
        raise NotImplementedError

    def save_telemetry(self, telemetry: List[Telemetry]):
        raise NotImplementedError

    def get_telemetry(self, exp_id=None) -> List[Telemetry]:
        raise NotImplementedError

    def save_event(self, event: ExperimentEvent):
        raise NotImplementedError

    def save_events(self, events: List[ExperimentEvent]):
        raise NotImplementedError

    def get_events(self, exp_id) -> List[ExperimentEvent]:
        raise NotImplementedError

    def save_nodeinfos(self, infos: List[NodeInfo]):
        raise NotImplementedError

    def find_all(self) -> List[Experiment]:
        raise NotImplementedError

    def finalize_experiment(self, exp: Experiment, status):
        exp.status = status
        exp.end = time.time()
        self.update_experiment(exp)
        self.touch_traces(exp)

    def get_running_experiment(self) -> Experiment:
        raise NotImplementedError



"""
Clients can publish into the redis topic 'galileo/events' with values: '<timestamp> <name> [<value>]'.
"""
import logging
import threading
from typing import Iterator

from galileodb import ExperimentDatabase
from galileodb.model import ExperimentEvent, Event

logger = logging.getLogger(__name__)


class RedisEventSubscriber:
    def __init__(self, rds, topic='galileo/events') -> None:
        super().__init__()
        self.rds = rds
        self.topic = topic
        self.pubsub = None

    def listen(self) -> Iterator[Event]:
        self.pubsub = self.rds.pubsub()

        try:
            self.pubsub.psubscribe(self.topic)

            for item in self.pubsub.listen():
                data = item['data']
                if type(data) == int:
                    continue

                # timestamp name [value]
                payload = data.split(' ', maxsplit=2)

                if len(payload) == 2 or len(payload) == 3:
                    yield Event(*payload)
                else:
                    logger.warning('Unknown event payload format %s', payload)

        finally:
            self.pubsub.close()

    def close(self):
        if self.pubsub:
            self.pubsub.punsubscribe()


class ExperimentEventRecorder:
    def __init__(self, rds, db: ExperimentDatabase, exp_id: str) -> None:
        self.rds = rds
        self.db = db
        self.exp_id = exp_id
        self._subscriber = None

    def run(self):
        self._subscriber = RedisEventSubscriber(self.rds)

        for event in self._subscriber.listen():
            try:
                self._record(event)
            except:
                logger.exception("error saving ExperimentEvent")

    def close(self):
        if self._subscriber:
            self._subscriber.close()

    def _record(self, event: Event):
        self.db.save_event(ExperimentEvent(self.exp_id, *event))


class BatchingExperimentEventRecorder:
    def __init__(self, rds, db: ExperimentDatabase, exp_id: str, flush_every=36) -> None:
        self.rds = rds
        self.db = db
        self.exp_id = exp_id
        self.flush_every = flush_every

        self._buffer = list()
        self._subscriber = None
        self._i = 0

    def run(self):
        self._subscriber = RedisEventSubscriber(self.rds)
        try:
            for event in self._subscriber.listen():
                self._record(event)
        finally:
            self.flush()

    def close(self):
        if self._subscriber:
            self._subscriber.close()

    def flush(self):
        if not self._buffer:
            logger.debug('event buffer empty')
            return

        logger.debug('saving %s event records of experiment "%s"', len(self._buffer), self.exp_id)

        self.db.save_events(self._buffer)
        self._buffer.clear()

    def _record(self, event: Event):
        self._buffer.append(ExperimentEvent(self.exp_id, *event))
        self._increment_and_flush()

    def _increment_and_flush(self):
        self._i = (self._i + 1) % self.flush_every
        if self._i == 0:
            self.flush()


class ExperimentEventRecorderThread(threading.Thread):
    def __init__(self, recorder, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.recorder = recorder

    def stop(self, timeout=None):
        self.recorder.close()
        self.join(timeout)

    def run(self) -> None:
        self.recorder.run()



import logging

from telemc import TelemetryRecorder, Telemetry, TelemetryController

from galileodb.model import Telemetry as GalileoTelemetry, NodeInfo
from galileodb.db import ExperimentDatabase

logger = logging.getLogger(__name__)


class ExperimentTelemetryRecorder(TelemetryRecorder):

    # TODO: need locks?

    def __init__(self, rds, db: ExperimentDatabase, exp_id: str, flush_every=36) -> None:
        super().__init__(rds)
        self.db = db
        self.exp_id = exp_id

        self.flush_every = flush_every
        self.i = 0
        self.buffer = list()

    def run(self):
        try:
            logger.debug('starting ExperimentTelemetryRecorder for experiment %s', self.exp_id)
            super().run()
        finally:
            logger.debug('closing ExperimentTelemetryRecorder for experiment %s', self.exp_id)
            self._flush()

    def save_nodeinfos(self):
        ctrl = TelemetryController(self.rds)
        infos = [NodeInfo(info.node, info.data, self.exp_id) for info in ctrl.get_node_infos()]
        logger.debug('saving node infos %s', infos)
        self.db.save_nodeinfos(infos)

    def _record(self, t: Telemetry):
        try:
            val = float(t.value)
        except ValueError:
            # the rationale is that this block will be executed rarely, and checking each time may be more expensive
            # than having the try/except block
            if t.metric == 'status':
                val = 1 if t.value == 'true' else 0
            else:
                logger.error('Could not convert value "%s" of metric "%s"', t.value, t.metric)
                return

        self.buffer.append(GalileoTelemetry(float(t.timestamp), t.metric, t.node, val, self.exp_id, t.subsystem))

        self.i = (self.i + 1) % self.flush_every
        if self.i == 0:
            self._flush()

    def _flush(self):
        if not self.buffer:
            return

        logger.debug('saving %s telemetry records of experiment "%s"', len(self.buffer), self.exp_id)

        self.db.save_telemetry(self.buffer)
        self.buffer.clear()
